/*
 *  gen05-pci.c
 *
 *  Based largely on:
 *  arch/mips/brcm/brcmpci.c -- Broadcom BCM93730 pci setup routines
 *
 *  Copyright (C) 2000 Troy Trammel <ttrammel@broadcom.com>
 *					Broadcom Corp.
 *
 *  Modifications Copyright (C) 2004 by TiVo, Inc.
 * 
 */

#include <linux/init.h>
#include <linux/kernel.h>
#define CONFIG_PCI 
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/ioport.h>

#include <asm/bcm7038.h>
#include <asm/io.h>

#include <asm/pci_channel.h>

#define	TIVO_VIXS_PCI_ID	0x21001745

#ifdef DUMP_PCI_CONFIG
static void PrintPCIDev( struct pci_dev *pci_dev )
{
    printk(
        "unsigned int   devfn                %08X  SLOT %02X FUNC %02X\n"
        "unsigned short vendor               %04X\n"
        "unsigned short device;              %04X\n"  
        "unsigned short subsystem_vendor     %04X\n" 
        "unsigned short subsystem_device     %04X\n"
        "unsigned int   class                %08X\n"
        "u8             hdr_type             %02X\n"
        "u8             rom_base_reg         %02X\n"
        "struct pci_driver *driver           %08X\n"
        "void           *driver_data         %08X\n"
        "dma_addr_t     dma_mask             %08X\n"
        "char           name[80]             %s\n"      
        "char           slot_name[8]         %s\n"
        "int            active               %08X\n"
        "int            ro                   %08X\n"
        "unsigned short regs                 %08X\n",
        pci_dev->devfn,  PCI_SLOT(pci_dev->devfn), PCI_FUNC(pci_dev->devfn),
        pci_dev->vendor,
        pci_dev->device,
        pci_dev->subsystem_vendor,
        pci_dev->subsystem_device,
        pci_dev->class,
        pci_dev->hdr_type,
        pci_dev->rom_base_reg,
        (int)pci_dev->driver,
        (int)pci_dev->driver_data,
        (int)pci_dev->dma_mask,
        pci_dev->name,
        pci_dev->slot_name,
        pci_dev->active,
        pci_dev->ro,
        pci_dev->regs);
}
#endif

#ifdef	CONFIG_MIPS_BCM7038C0
spinlock_t pciconf_lock = SPIN_LOCK_UNLOCKED;
static struct resource pci_io_resource = {
	"io pci IO space",
	0x0000000,
	0x0000000 + BCM_PCIIO_SIZE,
	IORESOURCE_IO,
	NULL, NULL, NULL
};

static struct resource pci_mem_resource = {
	"pci memory space",
	BCM_PCIMEM_PHYS,
	BCM_PCIMEM_PHYS+ BCM_PCIMEM_SIZE,
	IORESOURCE_MEM,
	NULL, NULL, NULL
};

static struct resource pci_sata_io_resource = {
	"io sata pci IO space",
	0,/*SATA_IO_BASE*/
	0 /*SATA_IO_BASE*/ + SATA_IO_SIZE,
	IORESOURCE_IO,
	NULL, NULL, NULL
};

static struct resource pci_sata_mem_resource = {
	"sata pci memory space",
	SATA_MEM_BASE,
	SATA_MEM_BASE+ SATA_MEM_SIZE,
	IORESOURCE_MEM,
	NULL, NULL, NULL
};

static int brcm_pci_read_config_dword( struct pci_dev *dev,
				int where, u32 *val)
{
    int slot_num, func_num;
    u32 cfc,cf8;
    unsigned long flags;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

    switch (dev->bus->number) {
    case 1: 
	/*
	 *On 7038 C0 SATA device in slot 0 is responding to all PCI discovery
	 *queries as if it were connected on all possible slots. (Bug???)
	 */ 
	if (dev->devfn != 0) 
	{
	    return PCIBIOS_FUNC_NOT_SUPPORTED;
	}
	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_DATA; 
    break;
    case 0:
	cfg_index_reg	=   (volatile u32 *)MIPS_PCI_XCFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)MIPS_PCI_XCFG_DATA; 
    break;
    default:

    /*
     *   configuration cycle 0 only (primary bus only)
     */
         printk("brcm_pci_write_config_dword: Invalid bus number\n");
	  return PCIBIOS_FUNC_NOT_SUPPORTED;
    }

    slot_num = PCI_SLOT(dev->devfn);	
    func_num = PCI_FUNC(dev->devfn);
    cf8 = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

    spin_lock_irqsave( &pciconf_lock, flags );

    *cfg_index_reg = cf8;
    cfc = *cfg_data_reg;
    *val = cfc;
    
    spin_unlock_irqrestore( &pciconf_lock, flags );
    
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_dword(struct pci_dev *dev, int where,
					u32 val)
{
    int slot_num, func_num;
    u32 cfc;
    unsigned long flags;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

    switch (dev->bus->number) {
    case 1: 
	/*
	 *On 7038 C0 SATA device in slot 0 is responding to all PCI discovery
	 *queries as if it were connected on all possible slots. (Bug???)
	 */ 
	if (dev->devfn != 0) 
	{
	    return PCIBIOS_FUNC_NOT_SUPPORTED;
	}
	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_DATA; 
    break;		
    case 0:
	cfg_index_reg	=   (volatile u32 *)MIPS_PCI_XCFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)MIPS_PCI_XCFG_DATA; 
    break;
    default:

    /*
     *   configuration cycle 0 only (primary bus only)
     */
       printk("brcm_pci_write_config_dword: Invalid bus number\n");
	return PCIBIOS_FUNC_NOT_SUPPORTED;
    }

    slot_num = PCI_SLOT(dev->devfn);
    func_num = PCI_FUNC(dev->devfn);

    cfc = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

    spin_lock_irqsave( &pciconf_lock, flags );

    *cfg_index_reg = cfc;
    *cfg_data_reg = val;

    spin_unlock_irqrestore( &pciconf_lock, flags );

    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	result >>= 16;
    *val = result & 0xffff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 1)
	result >>= 8;
    if (where & 2)
	result >>= 16;
    *val = result & 0xff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
    int status, shift = 0;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    result &= ~(0xffff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where & ~3, result);
}

static int brcm_pci_write_config_byte( struct pci_dev *dev, int where, u8 val)
{
    int status, shift = 0;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    if (where & 1)
	shift += 8;
    result &= ~(0xff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where & ~3, result);
}

struct pci_ops brcm_pci_ops = {
    brcm_pci_read_config_byte,
    brcm_pci_read_config_word,
    brcm_pci_read_config_dword,
    brcm_pci_write_config_byte,
    brcm_pci_write_config_word,
    brcm_pci_write_config_dword
};

struct pci_channel mips_pci_channels[] = {
	/* Primary PCI bus */
	{ &brcm_pci_ops, &pci_io_resource, &pci_mem_resource, 1, 0xff },

	/* Secondary PCI bus only contains 1 SATA controller at devfn 0 */
	{ &brcm_pci_ops, &pci_sata_io_resource, &pci_sata_mem_resource, 0, 0},
	{ NULL, NULL, NULL, 0, 0}
};

static void bcm_pcierr_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    printk( KERN_WARNING "PCI Error detected, IRQ=%d @0x%08lx\n", irq, regs->cp0_epc );
}

void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;
	int slot_num;
	int bus_num;

	pci_for_each_dev(dev) {

		slot_num = PCI_SLOT(dev->devfn);
		bus_num = dev->bus->number;
		switch (bus_num) 
		{
		    case 0:
			switch (slot_num) 
			{
			    case PCI_DEVICE_ID_7411_A: /* 7411 A Controller */

			      break;
			    case PCI_DEVICE_ID_7042_A:	/* 7042 A controller */
				    break;
			    case PCI_DEVICE_ID_7042_B:	/* 7042 B controller */
				    break;
			    case PCI_DEVICE_ID_EXT:
				    break;
			    default:
				    printk("unknown slot num 0x%x on bus 0x%x for pcibios_fixup_irqs\n", slot_num,bus_num);
			}
		    break;
		    case 1:
			switch (slot_num) 
			{
			    case PCI_DEVICE_ID_SATA:	/* IDE controller */
				    pcibios_update_irq(dev, BCM_IRQ_PCI_SATA_CPU_INTR);
				    dev->irq = BCM_IRQ_PCI_SATA_CPU_INTR;
				    break;
			    default:
				    printk("unknown slot num 0x%x on bus 0x%x for pcibios_fixup_irqs\n", slot_num,bus_num);
			}
		    break;
		    default:
			    printk("unknown bus num 0x%x for pcibios_fixup_irqs\n", bus_num);
		}
	}

	request_irq( BCM_IRQ_PCI_ADR_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SERR_DET_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_MSTR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_TAR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SRC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_DMA_ERROR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_RBUS_MSTR_ERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
}

// Remembers which devices have already been assigned.
static  int slots[32];

/* Do IO/memory allocation for expansion slot here */
static __init void brcm_pcibios_fixup_plugnplay(struct pci_dev *dev) 
{
	int i,j,k,l;
	volatile unsigned int iosize, msize;
	volatile u32 memAddr = PCI_EXPANSION_PHYS_MEM_WIN0_BASE;
	volatile u32 ioAddr = PCI_EXPANSION_PHYS_IO_WIN0_BASE;
	//int useIOAddr = 0;
	u32 savRegData;
	u32 size;
	u32 mask;
	int slot;
	u32 regData;
		
	/*
	* Make sure that it is in the PCI slot, otherwise we punt.
	*/

	slot = PCI_SLOT(dev->devfn);

	if (slots[slot]) {
		printk("Skip resource assignment for dev ID %04x:%04x already allocated\n", dev->vendor, dev->device);
		return;
	}

	slots[slot] = 1;

	pcibios_update_irq(dev, BCM_IRQ_PCI_INTA_2_CPU_INTR);
	dev->irq = BCM_IRQ_PCI_INTA_2_CPU_INTR;

	printk("\tPCI DEV in slot %x, ID=%04x:%04x\n\r", slot, dev->vendor, dev->device);

	/* Write FFFFFFFFH to BAR registers to probe for IO or Mem */
	iosize = msize = 0;
	for (i=0,j=0; i<6; i++,j+=4) {
		pci_write_config_dword(dev,PCI_BASE_ADDRESS_0+j,0xffffffff);
		pci_read_config_dword(dev,PCI_BASE_ADDRESS_0+j,&regData);
		if (regData) {
			printk("PCI PnP: PCI_BAR[%d] = %x\n", i, regData);

			savRegData = regData;
			regData &= 0xfffffff0; /* Get rid of Bit 0-3 0=Mem, 2:1 32/64, 3=Prefetchable */
			/*
			* determine io/mem size here by looking at the BAR
			* register, skipping the first 4 bits
			*/
			for (k=4; k<32; k++) {
				if (regData & (1<<k)) {
					break;
				}
			}
			if (k < 32) {
				printk("PCI PnP: size requested is %x\n", 1<<k);
				size = 1<<k;
				mask = 0xffffffff;
				for (l=0; l<k; l++) {
					mask &= ~(1<<l);
				}
			}
			else
				break;
			/* Only the first BAR counts on most devices */
			//if (i == 0) {
			//	useIOAddr = savRegData & 0x1;
			//}

			if (savRegData & 0x1) {	/* IO address */
				/* Calculate the next address that satisfies the boundary condition */
				regData = (ioAddr + size - 1) & mask;
				pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, regData);
				printk("PCI PnP: Writing PCI_IO_BAR[%d]=%x, size=%d, mask=%x\n", i,
					regData, 1<<k, mask);
				iosize = regData + size;
				ioAddr = regData + size; // Advance it
			}
			else { /* Mem address, tag on to the last one, the 7041 mem area */
				/* Calculate the next address that satisfies the boundary condition */
				regData = (memAddr + size - 1) & ~(size -1);
				pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, regData);
				printk("PCI PnP: Writing PCI_MEM_BAR[%d]=%x, size=%d, mask=%x\n", i,
					regData, 1<<k, mask);
				msize = regData + size;
				memAddr = regData+size; // Advance it
			}
		}
	}
	/*  now that it's all configured, turn on the Command Register */
	pci_read_config_dword(dev,PCI_COMMAND,&regData);
	regData |= PCI_COMMAND_SERR|PCI_COMMAND_PARITY |PCI_COMMAND_MASTER;
	if (iosize > 0)
		regData |= PCI_COMMAND_IO;
	if (msize > 0)
		regData |= PCI_COMMAND_MEMORY;

	printk("\tPCI PnP: PCI_DEV %04x:%04x, command=%x, memAddr=%08x, ioAddr=%08x\n", 
		dev->vendor, dev->device, regData, memAddr, ioAddr);
	pci_write_config_dword(dev,PCI_COMMAND,regData);
	
}
void __init pcibios_update_irq(struct pci_dev *dev, int irq)
{
    u8 val;

	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	/* Test this as it does not seem to be working */
	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &val);
}
unsigned __init int pcibios_assign_all_busses(void)
{
	return 0;
}
void __init pcibios_fixup_resources(struct pci_dev *dev)
{
    return;
}
void __init brcm_pcibios_fixup_device(struct pci_dev *dev)
{
	u32 dev_vendor_id;
	u32 regData;
	u32 size;

	/********************************************************
	 * Configure the Device
	 ********************************************************/
	dev_vendor_id = (u32)dev->device<<16;
	dev_vendor_id |= dev->vendor;

	switch (dev_vendor_id) {
	    /********************************************************
	     * SATA Controller
	     ********************************************************/
	    case BCM7038_SATA_VID:
	    {
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, SATA_PCS0_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, SATA_PCS1_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_2, SATA_SCS0_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_3, SATA_SCS1_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_4, SATA_CFG_BUS_MASTER_OFS+1);
		    
		    size = CPU2PCI_PCI_SATA_PHYS_MEM_WIN0_SIZE;
		    regData = CPU2PCI_PCI_SATA_PHYS_MEM_WIN0_BASE; 
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_5, regData);

		    regData = 0x0f;
		    pci_write_config_byte(dev, 0x3e, regData); /* Minimum Grant, to avoid Latency being reset to lower value */
		    
		    regData = 0xff;
		    pci_write_config_byte(dev, 0xd, regData); /* SATA_CFG_MASTER_LATENCY_TIMER. */


		    pci_read_config_dword(dev, PCI_COMMAND, &regData);
		    regData |= (PCI_COMMAND_SERR|PCI_COMMAND_PARITY|PCI_COMMAND_MASTER|PCI_COMMAND_IO| PCI_COMMAND_MEMORY);
		    pci_write_config_dword(dev, PCI_COMMAND, regData);

		    /* Primary */
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);

		    regData &= 0xfffffc; // Mask off  Reserved & Res Type bits.

		    regData += 2; // Offset 302H for primary
		    outb(inb(regData)|0x60,regData);

                    /* Secondary */
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);

		    regData &= 0xfffffc; // Mask off  Reserved & Res Type bits.

                    regData += 0xa; // Offset 30AH for secondary
                    outb(inb(regData)|0x60,regData);

	    }
	    break;

		case TIVO_VIXS_PCI_ID:
		{
			 u8 byteVal;

             printk("Found ViXS chip, dev_vendor_id = 0x%lx\n", dev_vendor_id);

             pcibios_update_irq(dev, BCM_IRQ_PCI_INTA_2_CPU_INTR);
             dev->irq = BCM_IRQ_PCI_INTA_2_CPU_INTR;
             printk(" ViXS dev-irq := %d\n", BCM_IRQ_PCI_INTA_2_CPU_INTR);

             pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,0xffffffff);
             pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &regData);
             printk(" ViXS BAR0 0x%x\n", regData);

	         pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,PCI_EXPANSION_PHYS_MEM_WIN0_BASE);
			 pci_write_config_dword(dev,PCI_BASE_ADDRESS_1,0xffffffff);
			 pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &regData);
			 printk(" ViXS BAR1 0x%x\n", regData);

			 pci_write_config_dword(dev,PCI_BASE_ADDRESS_1,PCI_EXPANSION_PHYS_MEM_WIN0_BASE+0x100000);
             pci_write_config_dword(dev,PCI_BASE_ADDRESS_2,0xffffffff);
             pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &regData);
             printk(" ViXS BAR2 0x%x\n", regData);

             pci_write_config_dword(dev,PCI_BASE_ADDRESS_3,0xffffffff);
             pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &regData);
             printk(" ViXS BAR3 0x%x\n", regData);

             pci_write_config_dword(dev,PCI_BASE_ADDRESS_4,0xffffffff);
             pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);
             printk(" ViXS BAR4 0x%x\n", regData);

             pci_write_config_dword(dev,PCI_BASE_ADDRESS_5,0xffffffff);
             pci_read_config_dword(dev, PCI_BASE_ADDRESS_5, &regData);
             printk(" ViXS BAR5 0x%x\n", regData);

             pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &byteVal);
             printk(" ViXS PCI interrupt line = 0x%x\n", byteVal);

             pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &byteVal);
             printk(" ViXS PCI interrupt pin = 0x%x\n", byteVal);

             pci_read_config_dword(dev,PCI_COMMAND,&regData);
             regData |= PCI_COMMAND_SERR|PCI_COMMAND_PARITY |PCI_COMMAND_MASTER;
             regData |= PCI_COMMAND_MEMORY;
             pci_write_config_dword(dev,PCI_COMMAND,regData);
             printk(" ViXS PCI_COMMAND := 0x%x\n", regData);


		}
		break;

	    /********************************************************
	     * Config the BCM7042 chip
	     ********************************************************/
	    case BCM7042_VID:
		    regData = BCM7042_MEM_BASE;
            if ( PCI_SLOT(dev->devfn) & 1 ) {
                regData += 0x400000; 
            }
		    printk("\tBCM7042 MPEG2 Encoder on PCI slot 0x%x @ %p\n", PCI_SLOT(dev->devfn), (void *)regData);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, PCI_UNUSED_MEM_ADDR);   // pre-fetchable 16MB
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, PCI_UNUSED_MEM_ADDR);   // pre-fetchable 16MB
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_2, PCI_UNUSED_MEM_ADDR);   // pre-fetchable 16MB
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_3, regData);               // real reg space 4MB
     
		    pci_read_config_dword(dev, PCI_COMMAND, &regData);
		    regData |= PCI_COMMAND_MEMORY;
		    pci_write_config_dword(dev, PCI_COMMAND, regData);
		    break;

	    /********************************************************
	     * Config the BCM7411 chip
	     ********************************************************/
	    case BCM7411_VID:
		    printk("\tBCM7411 Decoder on PCI slot 0x%x @ %p\n", PCI_SLOT(dev->devfn), (void *)BCM7411_MEM_BASE);
		    regData = BCM7411_MEM_BASE;
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, regData);   // 16MB
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, regData+0x1000000);   //  4K
		    pci_read_config_dword(dev, PCI_COMMAND, &regData);
		    regData |= PCI_COMMAND_MEMORY;
		    pci_write_config_dword(dev, PCI_COMMAND, regData);

		    printk("\t  cmd 0x%x", regData);
		    pci_read_config_dword(dev, PCI_COMMAND, &regData);
		    printk(" 0x%x", regData);
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &regData);
		    printk(" BAR0 0x%x", regData);
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &regData);
		    printk(" BAR1 0x%x\n", regData);

		    break;
	    default:
		    /* FOr 7038C0, it is not really Expansion slot any more, we just let the Plug&Play logic does the assignment */
		    printk("unknown dev_man id 0x%x, trying expansion slot\n", dev_vendor_id);
		    brcm_pcibios_fixup_plugnplay(dev);
	    break;
	}
#ifdef DUMP_PCI_CONFIG
    PrintPCIDev( dev );
	{
	    int i;
	    printk("\tPCI Config Space IDSEL%d",PCI_SLOT(dev->devfn));
	    for (i=0;i<0x40;i++) {
		    if ((i & 3) == 0) printk("\r\n\t\t");
		    pci_read_config_byte(dev,i,(unsigned char *)&regData);
		    printk("[%.2x]=%.2x  ",i,regData & 0xFF);
	    }
	    printk("\n\r");
	}
#endif
	return;
}

void __init pcibios_fixup(void)
{
    struct pci_dev *dev;
    pci_for_each_dev(dev) {
		/* <TDT> Our enable_device sets up bios data for the bootrom does not */
		brcm_pcibios_fixup_device(dev);

		/* <TDT> force a rescan in case we still need bios data */
		pci_setup_device(dev);
    }
}
void pcibios_set_master( struct pci_dev *dev )
{
    u16 command;
    unsigned long flags;
 
    save_and_cli(flags);
    pci_read_config_word(dev, PCI_COMMAND, &command);
    command |= PCI_COMMAND_MASTER;
    pci_write_config_word(dev, PCI_COMMAND, command);
    restore_flags(flags);
}
#if 0
/* --------------------------------------------------------------------------
    Name: DumpRegs
Abstract: Dump the IDE registers
 -------------------------------------------------------------------------- */

void
dump_ide(void)
{
	
	volatile unsigned long mmioBase = SATA_MEM_BASE;
	u16 cmd, status;
	u32 sataIntStatus;
	struct pci_dev *dev;
    

	printk("7038B0 SATA: Status Reg = %08x\n", *((volatile unsigned long *) (mmioBase+0x1C)));
	printk("7038B0 SATA: Port0 Master Status = %08x\n", *((volatile unsigned long *) (mmioBase+0x30)));
	printk("7038B0 SATA: Port0 SATA Status = %08x\n", *((volatile unsigned long *)  (mmioBase+0x40)));
	printk("7038B0 SATA: Port0 SATA Error = %08x\n", *((volatile unsigned long *)  (mmioBase+0x44)));

       pci_for_each_dev(dev) {
	   	u32 dev_vendor_id;
	
		dev_vendor_id = (u32)dev->device<<16;
		dev_vendor_id |= dev->vendor;

		switch (dev_vendor_id) {
	
		case BCM7038_SATA_VID:
	   	
			pci_read_config_word(dev, PCI_COMMAND, &cmd);
			pci_read_config_word(dev, PCI_STATUS, &status);
			pci_read_config_dword(dev, 0x80, &sataIntStatus);
			printk("7038B0 SATA: PCI Command=%04x, Status = %04x, sataIntStatus=%08x\n", cmd, status, sataIntStatus);
			break;
		default:

		}
       	}
	
}
EXPORT_SYMBOL(dump_ide);
#endif
#else //CONFIG_MIPS_BCM7038C0
spinlock_t pciconf_lock = SPIN_LOCK_UNLOCKED;

static int brcm_pci_read_config_dword( struct pci_dev *dev, int where, u32 *val)
{
    unsigned long flags;
    int slot_num, func_num;

    /*
     *  For starters let's do configuration cycle 0 only (one bus only)
     */
    if (dev->bus->number)
	return PCIBIOS_FUNC_NOT_SUPPORTED;

    slot_num = PCI_SLOT(dev->devfn);
    func_num = PCI_FUNC(dev->devfn);
    
    spin_lock_irqsave( &pciconf_lock, flags );
    *(volatile u32*)(MIPS_PCI_XCFG_INDEX) = 
        PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;
    barrier();
    
    *val = *(volatile u32*)(MIPS_PCI_XCFG_DATA);
    barrier();
    spin_unlock_irqrestore( &pciconf_lock, flags );
    
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_dword(struct pci_dev *dev, int where, u32 val)
{ 
    unsigned long flags;
    int slot_num, func_num;

    /*
     *  For starters let's do configuration cycle 0 only (one bus only)
     */
    if (dev->bus->number)
	return PCIBIOS_FUNC_NOT_SUPPORTED;

    slot_num = PCI_SLOT(dev->devfn);
    func_num = PCI_FUNC(dev->devfn);

    
    spin_lock_irqsave( &pciconf_lock, flags );
    *(volatile u32*)(MIPS_PCI_XCFG_INDEX) = 
        PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;
    barrier();
    *(volatile u32*)(MIPS_PCI_XCFG_DATA) = val;
    barrier();
    spin_unlock_irqrestore( &pciconf_lock, flags );

    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	result >>= 16;
    *val = result & 0xffff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 1)
	result >>= 8;
    if (where & 2)
	result >>= 16;
    *val = result & 0xff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
    int status, shift = 0;
    u32 result;
    status = brcm_pci_read_config_dword(dev, where, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    result &= ~(0xffff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where, result);
}

static int brcm_pci_write_config_byte( struct pci_dev *dev, int where, u8 val)
{
    int status, shift = 0;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    if (where & 1)
	shift += 8;
    result &= ~(0xff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where, result);
}

struct pci_ops brcm_pci_ops = {
    brcm_pci_read_config_byte,
    brcm_pci_read_config_word,
    brcm_pci_read_config_dword,
    brcm_pci_write_config_byte,
    brcm_pci_write_config_word,
    brcm_pci_write_config_dword
};

/* Do IO/memory allocation for expansion slot here, merged from 2.4.25 implementation */
static __init void brcm_pcibios_fixup_expansion_slot(struct pci_dev *dev) 
{
	int i,j,k,l, iosize, msize;
	u32 memAddr = PCI_EXPANSION_PHYS_MEM_WIN0_BASE;
	u32 ioAddr = PCI_EXPANSION_PHYS_IO_WIN0_BASE;
	//int useIOAddr = 0;
	u32 savRegData;
	u32 size;
	u32 mask;
	int slot;
	u32 regData;
		

	/*
	* Make sure that it is in the PCI slot, otherwise we punt.
	*/


	slot = PCI_SLOT(dev->devfn);

	/* Only handle devices in the PCI slot */
	if ( PCI_DEVICE_ID_EXT != slot) {
		printk(KERN_WARNING "\tDev ID %04x:%04x ignored\n", dev->vendor, dev->device);
		return;
	}

//	pcibios_update_irq(dev, BCM_LINUX_EXT_PCI_IRQ);
//	dev->irq = BCM_LINUX_EXT_PCI_IRQ;


	printk(KERN_WARNING "\tPCI DEV in PCI slot, ID=%04x:%04x\n\r", dev->vendor, dev->device);
	

	/* Write FFFFFFFFH to BAR registers to probe for IO or Mem */
	iosize = msize = 0;
	for (i=0,j=0; i<6; i++,j+=4) {
		pci_write_config_dword(dev,PCI_BASE_ADDRESS_0+j,0xffffffff);
		pci_read_config_dword(dev,PCI_BASE_ADDRESS_0+j,&regData);
		if (regData) {
			printk(KERN_WARNING "Expansion: PCI_BAR[%d] = %x\n", i, regData);


			savRegData = regData;
			regData &= 0xfffffff0; /* Get rid of Bit 0-3 0=Mem, 2:1 32/64, 3=
Prefetchable */
			/*
			* determine io/mem size here by looking at the BAR
			* register, skipping the first 4 bits
			*/
			for (k=4; k<32; k++) {
				if (regData & (1<<k)) {
					break;
				}
			}
			if (k < 32) {
				printk("Expansion: size requested is %x\n", 1<<k);
				size = 1<<k;
				mask = 0xffffffff;
				for (l=0; l<k; l++) {
					mask &= ~(1<<l);
				}
			}
			else
				break;
			/* Only the first BAR counts on most devices */
			//if (i == 0) {
			//	useIOAddr = savRegData & 0x1;
			//}

			if (savRegData & 0x1) {	/* IO address */
				/* Calculate the next address that satisfies the boundary condition */
				ioAddr = (ioAddr + size - 1) & mask;
				pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, ioAddr);
				printk(KERN_WARNING "Expansion: Writing PCI_IO_BAR[%d]=%x, size=%d, mask=%x\n", i,
					ioAddr, 1<<k, mask);
				iosize += size;
			}
			else { /* Mem address, tag on to the last one, the SATA mem area */
				/* Calculate the next address that satisfies the alignment condition */
				memAddr = (memAddr + size - 1) & ~(size -1);
				pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, memAddr);
				printk(KERN_WARNING "Expansion: Writing PCI_MEM_BAR[%d]=%x, size=%d, mask=%x\n", i,
					memAddr, 1<<k, mask);
				msize = memAddr + size;
			}
		}
	}
	/*  now that it's all configured, turn on the Command Register */
	pci_read_config_dword(dev,PCI_COMMAND,&regData);
	regData |= PCI_COMMAND_SERR|PCI_COMMAND_PARITY |PCI_COMMAND_MASTER;
	if (iosize > 0)
		regData |= PCI_COMMAND_IO;
	if (msize > 0)
		regData |= PCI_COMMAND_MEMORY;

	printk( KERN_WARNING "\tExpansion: PCI_DEV %04x:%04x, command=%x\n", 
		dev->vendor, dev->device, regData);
	pci_write_config_dword(dev,PCI_COMMAND,regData);
}
static int pcibios_configure_device(struct pci_dev *dev, int mask )
{
	u32 dev_vendor_id;
	u32 regData, size;

	/********************************************************
	 * Configure the Device
	 ********************************************************/
	dev_vendor_id = (u32)dev->device<<16;
	dev_vendor_id |= dev->vendor;
	printk("dev_vendor_id %x\n",dev_vendor_id);
	switch (dev_vendor_id) {
	/********************************************************
	 * SATA Controller
	 ********************************************************/
	case BCM7038_SATA_VID:
		printk( "\tBCM7038 IDE/Serial ATA controller\n" );
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, SATA_PCS0_OFS+1 );
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, SATA_PCS1_OFS+1 );
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_2, SATA_SCS0_OFS+1 );
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_3, SATA_SCS1_OFS+1 );
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_4, SATA_CFG_BUS_MASTER_OFS+1);
#ifdef CONFIG_MIPS_BCM7038B0
		/* Turn on BAR 5 in order to set PLL */
 		size = PCI_SATA_PHYS_MEM_WIN5_SIZE;
		regData = (PCI_SATA_PHYS_MEM_WIN5_BASE + size -1) & ~(size-1);
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_5, regData);
		pci_write_config_byte(dev, 0xd, 0xff); /* SATA_CFG_MASTER_LATENCY_TIMER. */

		/*  Turn on MEM */
		pci_read_config_dword(dev, PCI_COMMAND, &regData);
		regData |= (PCI_COMMAND_MASTER|PCI_COMMAND_IO| PCI_COMMAND_MEMORY);
		pci_write_config_dword(dev, PCI_COMMAND, regData);

		pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);
		regData &= 0xfffffc; // Mask off  Reserved & Res Type bits.
#ifdef SATA_DEBUG
		printk("SATA: Bus Master Status Register offset = %08lx + %08x = %08x\n", 
			mips_io_port_base, regData, mips_io_port_base+regData);
		printk("SATA: before init Bus Master Status reg = 0x%08x.\n", *((volatile unsigned char *)(mips_io_port_base+regData)));
#endif
		regData += 2; // Offset 302H
		*((volatile unsigned char *)(mips_io_port_base+regData)) |= 0x20;
#ifdef SATA_DEBUG
		printk("SATA: after init Bus Master Status reg = 0x%08x.\n", *((volatile unsigned char *)(mips_io_port_base+regData)));
#endif

		// Force PLLLOCK
		*((volatile unsigned long *)(PCI_SATA_PHYS_MEM_WIN5_BASE+0x84)) |= 0x08000000;

#ifdef SATA_DEBUG
		printk("SATA: after PLLLOCK reg = 0x%08lx.\n", *((volatile unsigned long *)(PCI_SATA_PHYS_MEM_WIN5_BASE+0x84)));

#endif
		// Turn off SATA 2
		*((volatile unsigned long *)(PCI_SATA_PHYS_MEM_WIN5_BASE+0x148)) |= 0x00000001;
#ifdef SATA_DEBUG
		printk("SATA: after PLLLOCK reg = 0x%08lx.\n", *((volatile unsigned long *)(PCI_SATA_PHYS_MEM_WIN5_BASE+0x148)));
#endif
#endif /* CONFIG_MIPS_BCM7038B0 */
		break;

	/********************************************************
	 * Config the BCM7041 chip
	 ********************************************************/
	case BCM7041_VID:
		printk("\tBCM7041 Dual Channel MPEG2 Encoder\n\r");
		regData = BCM7041_MEM_BASE;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, regData);
		regData += 0x20000;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, regData);
		regData += 0x20000;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_2, regData);
		regData += 0x20000;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_3, regData);
		regData += 0x20000;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_4, regData);
		regData += 0x20000;
		pci_write_config_dword(dev, PCI_BASE_ADDRESS_5, regData);
 
		pci_read_config_dword(dev, PCI_COMMAND, &regData);
		regData |= PCI_COMMAND_MEMORY;
		pci_write_config_dword(dev, PCI_COMMAND, regData);
		break;

	default:
		//printk("unknown dev_man id 0x%x for pcibios_enable_device\n", dev_vendor_id);
		brcm_pcibios_fixup_expansion_slot(dev);
		break;
	}
#ifdef DUMP_PCI_CONFIG
	{
	int i;
	printk( KERN_WARNING "\tPCI Config Space IDSEL%d"),PCI_SLOT(dev->devfn));
	for (i=0;i<0x40;i++) {
		if ((i & 3) == 0) printk("\r\n\t\t");
		pci_read_config_byte(dev,i,(unsigned char *)&regData);
		printk("[%.2x]=%.2x  "),i,regData & 0xFF);
	}
	printk("\n\r");
	}
#endif
	return 0;
}

/*
 * We enable all devices during setup as the ALi must be setup in order
 * to access devices which use it's PIC.
 */
static void __init pcibios_enable_devices(void)
{
    extern int pcibios_enable_device( struct pci_dev *, int );

    struct pci_dev *dev;
    pci_for_each_dev(dev) {
		printk("Enabling PCI devices\n");
		/* <TDT> Our enable_device sets up bios data for the bootrom does not */
		pcibios_configure_device(dev, 0);

		/* <TDT> force a rescan in case we still need bios data */
		pci_setup_device(dev);
    }
}

/*
 * Our current boot rom does not set these properly, so we change
 * them during setup.
 */
static void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;
	int slot_num;

	pci_for_each_dev(dev) {
		slot_num = PCI_SLOT(dev->devfn);
		switch (slot_num) {
		case PCI_DEVICE_ID_3250:		/* BCM3250  */
			pcibios_update_irq(dev, BCM_IRQ_3250_CPU_INTR);
			dev->irq = BCM_IRQ_3250_CPU_INTR;
			break;
		case PCI_DEVICE_ID_SATA:	/* IDE controller */
			pcibios_update_irq(dev, BCM_IRQ_PCI_SATA_CPU_INTR);
			dev->irq = BCM_IRQ_PCI_SATA_CPU_INTR;
			break;
		case PCI_DEVICE_ID_EXT:	/* On-board PCI slot */
			pcibios_update_irq(dev, BCM_IRQ_EXT_IRQ_0_CPU_INTR);
			dev->irq = BCM_IRQ_EXT_IRQ_0_CPU_INTR;
			break;
		case PCI_DEVICE_ID_7041:
			/* do not need to fixup. */
		break;
		
		default:
			printk("unknown slot num 0x%x for pcibios_fixup_irqs\n", slot_num);
		break;
		}
	}
}

static void bcm_pcierr_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    printk( KERN_WARNING "PCI Error detected, IRQ=%d @0x%08lx\n", irq, regs->cp0_epc );
}

void __init pcibios_init(void)
{
	printk( KERN_WARNING "PCI: Probing PCI hardware\n");

//	ioport_resource.end = BCM_PCIIO_SIZE - 1;
//	iomem_resource.start = BCM_PCIMEM_PHYS;
//	iomem_resource.end   = BCM_PCIMEM_PHYS + BCM_PCIMEM_SIZE - 1;
	
	printk("About to scan all bus 0\n");
	pci_scan_bus(0, &brcm_pci_ops, NULL);
	pcibios_enable_devices();
	pcibios_fixup_irqs();

	request_irq( BCM_IRQ_PCI_ADR_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SERR_DET_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_MSTR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_TAR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SRC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_DMA_ERROR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_RBUS_MSTR_ERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
}


char *pcibios_setup (char *str)
{
    return str;
}

void __init pcibios_update_irq(struct pci_dev *dev, int irq)
{
    u8 val;

	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	/* Test this as it does not seem to be working */
	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &val);
}


int pcibios_enable_resources(struct pci_dev *dev)
{
	u16 cmd, old_cmd;
	int idx;
	struct resource *r;
	printk(KERN_WARNING "PCI: Enabling device %x\n", dev->device);

	pci_read_config_word(dev, PCI_COMMAND, &cmd);
	old_cmd = cmd;
	for(idx=0; idx<6; idx++) {
		r = &dev->resource[idx];
		if (!r->start && r->end) {
			printk(KERN_ERR "PCI: Device %s not available because of resource collisions\n", dev->slot_name);
			printk(KERN_ERR "r->start=%lx, r->end=%lx\n",r->start,r->end);
			return -EINVAL;
		}
		if (r->flags & IORESOURCE_IO)
			cmd |= PCI_COMMAND_IO;
		if (r->flags & IORESOURCE_MEM)
			cmd |= PCI_COMMAND_MEMORY;
	}
	if (cmd != old_cmd) {
		printk(KERN_WARNING "PCI: Enabling device %s (%04x -> %04x)\n", dev->slot_name, old_cmd, cmd);
		pci_write_config_word(dev, PCI_COMMAND, cmd);
	}
	return 0;
}


int pcibios_enable_device(struct pci_dev *dev, int mask )
{
  return PCIBIOS_SUCCESSFUL;
}

void pcibios_set_master( struct pci_dev *dev )
{
    u16 command;
    unsigned long flags;
 
    save_and_cli(flags);
    pci_read_config_word(dev, PCI_COMMAND, &command);
    command |= PCI_COMMAND_MASTER;
    pci_write_config_word(dev, PCI_COMMAND, command);
    restore_flags(flags);
}

void pcibios_update_resource(struct pci_dev *dev, struct resource *root,
			     struct resource *res, int resource)
{
}

void __init pcibios_fixup_bus(struct pci_bus *bus)
{
}

void pcibios_align_resource( void *data, struct resource *res,
			    unsigned long size, unsigned long something )
{
  
}

unsigned __init int pcibios_assign_all_busses(void)
{
	return 0;
}

struct pci_fixup pcibios_fixups[] = { {0} };
#endif
